<?php

class RedisQueue implements DrupalQueueInterface {
  // The name of the queue that holds available items.
  protected $avail;
  // The name of the queue that holds claimed items.
  protected $claimed;
  // The PhpRedis object which connects to the redis server.
  protected $queue;

  protected $reserve_timeout;

  /**
   * Start working with a queue.
   *
   * @param $name
   *   Arbitrary string. The name of the queue to work with.
   */
  public function __construct($name) {
    $this->avail= 'drupal:queue:' . $name . ':avail';
    $this->claimed = 'drupal:queue:' . $name . ':claimed';
    $this->queue = new Redis();
    $options = self::getOptions($name);
    $this->reserve_timeout = $options['reserve_timeout'];
    $this->queue->connect($options['host'], $options['port']);
  }

  static function getOptions($name) {
    $options = variable_get('redis_queue_' . $name, array());
    $defaults = variable_get('redis_default_queue', array()) + array(
      'host' => variable_get('redis_client_host', '127.0.0.1'),
      'port' => variable_get('redis_client_port', 6379),
      'reserve_timeout' => NULL,
    );
    $options += $defaults;

    return $options;
  }

  /**
   * Add a queue item and store it directly to the queue.
   *
   * @param $data
   *   Arbitrary data to be associated with the new task in the queue.
   * @return
   *   TRUE if the item was successfully created and was (best effort) added
   *   to the queue, otherwise FALSE. We don't guarantee the item was
   *   committed to disk, that your disk wasn't hit by a meteor, etc, but as
   *   far as we know, the item is now in the queue.
   */
  public function createItem($data) {
    $record = new stdClass();
    $record->data = $data;
    $record->qid = $this->incrementId();
    $record->timestamp = time();
    $record = serialize($record);
    return $this->queue->lpush($this->avail, $record) ? TRUE : FALSE;
  }

  protected function incrementId() {
    return $this->queue->incr($this->avail . '_counter');
  }

  /**
   * Retrieve the number of items in the queue.
   *
   * This is intended to provide a "best guess" count of the number of items in
   * the queue. Depending on the implementation and the setup, the accuracy of
   * the results of this function may vary.
   *
   * e.g. On a busy system with a large number of consumers and items, the
   * result might only be valid for a fraction of a second and not provide an
   * accurate representation.
   *
   * @return
   *   An integer estimate of the number of items in the queue.
   */
  public function numberOfItems() {
    return $this->queue->lLen($this->avail);
  }

  /**
   * Claim an item in the queue for processing.
   *
   * @param $lease_time
   *   How long the processing is expected to take in seconds, defaults to an
   *   hour. After this lease expires, the item will be reset and another
   *   consumer can claim the item. For idempotent tasks (which can be run
   *   multiple times without side effects), shorter lease times would result
   *   in lower latency in case a consumer fails. For tasks that should not be
   *   run more than once (non-idempotent), a larger lease time will make it
   *   more rare for a given task to run multiple times in cases of failure,
   *   at the cost of higher latency.
   * @return
   *   On success we return an item object. If the queue is unable to claim an
   *   item it returns false. This implies a best effort to retrieve an item
   *   and either the queue is empty or there is some other non-recoverable
   *   problem.
   */
  public function claimItem($lease_time = 3600) {
    if (is_numeric($this->reserve_timeout)) {
      return $this->claimItemBlocking($lease_time);
    }
    else {
      $job = $this->queue->rpop($this->avail);
      $this->queue->zAdd($this->claimed, time() + $lease_time, $job);
      if ($job) {
        $item = unserialize($job);
        return $item;
      }
      return FALSE;
    }
  }

  /**
   * A blocking version of claimItem to be used with long-running queue workers
   * like waiting_queue.
   */
  public function claimItemBlocking($lease_time) {
    $job = $this->queue->brpop($this->avail, $this->reserve_timeout);
    $this->queue->zAdd($this->claimed, time() + $lease_time, $job);
    if ($job) {
      $item = unserialize($job[1]);
      return $item;
    }
    return FALSE;
  }

  /**
   * Delete a finished item from the queue.
   *
   * @param $item
   *   The item returned by DrupalQueueInterface::claimItem().
   */
  public function deleteItem($item) {
    return $this->queue->lrem($this->claimed, $item, -1);
  }

  public function createQueue() {}

  /**
   * Delete a queue and every item in the queue.
   */
  public function deleteQueue() {
    $this->queue->del($this->avail, $this->claimed, $this->claimed_counter);
  }

  public function releaseItem($item) {
    $res = $this->queue->multi()->zrem($this->claimed, $item)->lpush($this->avail, serialize($item))->exec();
  }


  /**
   * Expire claims in all queues.
   *
   * @return
   */
  public function expire() {
    foreach ($this->queue->keys("*:claimed") as $name) {
      $items = $this->queue->zRangeByScore($name, 0, time());
      foreach ($items as $item) {
        $this->releaseItem($item);
      }
    }
  }

  /**
   * Dumps items in the queue.
   *
   * @return
   *   An array of queue items.
   */
  public function dump() {
    return $this->queue->lrange($this->avail, 0, -1);
  }

  /**
   * Dump currently claimed queue items.
   *
   * @return
   *   An array of queue items.
   */
  public function dumpClaimed() {
    return $this->queue->zRevRange($this->claimed, 0, -1, TRUE);
  }
}
